# -*- coding: latin-1 -*-
#########################################################
#Author: Jean-Baptiste Fasquel, ISTIA, Angers University, France
#Copyright (C) 2014 Jean-Baptiste Fasquel
#########################################################
import numpy as np
import cv2

def create_K(width_current=None,height_current=None):
    """
    if not None width_current, height_current are used to adapt
    K with respect to, for instance, a current image the size of
    which would differ from the reference image (width_ref,height_ref) used for calibration
    (e.g. smaller size)
    create_K(image_current.shape[1],image_current.shape[0])
    """
    #SIZE OF REFERENCE IMAGE
    width_ref=960
    height_ref=540
    #CONNAISSANCE DE LA GEOMETRIE D'ACQUISITION
    dX=155.0 #mm
    dY=105.0 #mm
    dZ=500.0 #mm
    #CONNAISSANCE DES DIMENSIONS EN PIXELS DANS L'IMAGE NUMERIQUE
    dx=243.0 #pixels
    dy=170.0 #pixels
    #CALCULS
    x_0=width_ref/2.0
    y_0=height_ref/2.0
    alpha_x=dx*dZ/dX
    alpha_y=dy*dZ/dY
    #IF CURRENT IMAGE HAS A DIFFERENT SIZE
    if width_current is not None:
        x_0=width_current/2.0
        alpha_x=alpha_x*width_current/float(width_ref)
    if height_current is not None:
        y_0=height_current/2.0
        alpha_y=alpha_y*height_current/float(height_ref)
    #MATRICE INTRINSEQUE
    K=np.zeros((3,3))
    K[0,0]=alpha_x
    K[1,1]=alpha_y
    K[0,2]=x_0
    K[1,2]=y_0
    K[2,2]=1
    return K

def pixels2camera(pixel_coords,K,Z):
    '''
    Without case "len(pixel_coords.shape) == 1", one must
    invoke: one_point_3d=helper.pixels2camera(np.array([one_point_2d]),K,Z=10)[0]
    '''
    #Case list of pixels (one pixel per row)
    if len(pixel_coords.shape) == 2 :
        #ETAPE 1: (x,y) -> (x,y,1) (coordonnees homogenes)
        vecteur_colonne_unitaire=np.ones((pixel_coords.shape[0],1)) #Vecteur colonne de 1
        pixel_coords_h=np.hstack((pixel_coords,vecteur_colonne_unitaire)) #Ajout de la colonne
        #ETAPE 2: (x,y,1) -> Z*(x,y,1)=(Zx,Zy,Z)=(x',y',z') 
        pixel_coords_h_Z=Z*pixel_coords_h
        #ETAPE 3: inversion de K
        K_inv=np.linalg.inv(K)
        #ETAPE 4: produit matriciel
        cam_coords=np.dot(K_inv,pixel_coords_h_Z.T).T
    #Case list of pixels (one pixel per row)
    elif len(pixel_coords.shape) == 1 :
        #ETAPE 1: (x,y) -> (x,y,1) (coordonnees homogenes)
        vecteur_colonne_unitaire=np.ones((1)) #Vecteur colonne de 1
        pixel_coords_h=np.hstack((pixel_coords,vecteur_colonne_unitaire)) #Ajout de la colonne
        #ETAPE 2: (x,y,1) -> Z*(x,y,1)=(Zx,Zy,Z)=(x',y',z') 
        pixel_coords_h_Z=Z*pixel_coords_h
        #ETAPE 3: inversion de K
        K_inv=np.linalg.inv(K)
        #ETAPE 4: produit matriciel
        cam_coords=np.dot(K_inv,pixel_coords_h_Z)
         
    return cam_coords

def world2pixels(world_coords,proj_matrix):
    if len(world_coords.shape) == 1 :
        #ETAPE 1: projection par produit matriciel -> coordonnees homogenes
        pixel_coords_h=np.dot(proj_matrix,world_coords.T).T
        #ETAPE 2: normalisation, pour chaque ligne (point x,y,z), on divise par le dernier terme (z) 
        #Remarque: on travaille sur l'ensemble des points (lignes) a chaque etape (optimisation)
        pixel_coords_h[0]=pixel_coords_h[0]/pixel_coords_h[2]
        pixel_coords_h[1]=pixel_coords_h[1]/pixel_coords_h[2]
        pixel_coords_h[2]=pixel_coords_h[2]/pixel_coords_h[2]
        #ETAPE 3: on ne conserve que les 2 premieres coordonnees (x et y)
        pixel_coords=pixel_coords_h[0:2].astype(np.float32) #float32 sinon probleme avec drawChessboardCorners
    elif len(world_coords.shape) == 2 :
        #ETAPE 1: projection par produit matriciel -> coordonnees homogenes
        pixel_coords_h=np.dot(proj_matrix,world_coords.T).T
        #ETAPE 2: normalisation, pour chaque ligne (point x,y,z), on divise par le dernier terme (z) 
        #Remarque: on travaille sur l'ensemble des points (lignes) a chaque etape (optimisation)
        pixel_coords_h[:,0]=pixel_coords_h[:,0]/pixel_coords_h[:,2]
        pixel_coords_h[:,1]=pixel_coords_h[:,1]/pixel_coords_h[:,2]
        pixel_coords_h[:,2]=pixel_coords_h[:,2]/pixel_coords_h[:,2]
        #ETAPE 3: on ne conserve que les 2 premieres coordonnees (x et y)
        pixel_coords=pixel_coords_h[:,0:2].astype(np.float32) #float32 sinon probleme avec drawChessboardCorners
    return pixel_coords

def create_square(world_coords,size):
    center=np.mean(world_coords,0)
    square = np.float32([[center[0]-0.5*size, center[1]-0.5*size,center[2]], 
                         [center[0]+0.5*size, center[1]-0.5*size,center[2]], 
                         [center[0]+0.5*size, center[1]+0.5*size,center[2]], 
                         [center[0]-0.5*size, center[1]+0.5*size,center[2]]])
    return square

def coord2homogeneous(coords):
    """
    add new column of ones
    """
    new_col=np.ones((coords.shape[0],1))
    coords_h=np.hstack((coords,new_col))
    return coords_h

def compute_extrinsic_matrix(pts3d,pts2d,K,verbose=False):
    dist_coef = np.zeros(4)
    retval,rvec,tvec=cv2.solvePnP(pts3d, pts2d, K, dist_coef)
    R,b=cv2.Rodrigues(rvec)
    M_ext=np.hstack((R,tvec))
    if verbose:
        print ("Rotation x (deg): " , np.round(rvec[0]*360.0/(2*np.pi),2))
        print ("Rotation y (deg): " , np.round(rvec[1]*360.0/(2*np.pi),2))
        print ("Rotation z (deg): " , np.round(rvec[2]*360.0/(2*np.pi),2))
        print ("Translation: ",np.round(tvec[0],2), ", " , np.round(tvec[1],2), ", "  , np.round(tvec[2],2))
        
        np.set_printoptions(suppress=True)
        print( "Matrice:\n", np.round(M_ext,2))
    return M_ext

def compute_extrinsic_matrix_ransac(pts3d,pts2d,K,verbose=False):
    #Opencv "ransac" waits for float 32 type and not float 64
    pts3d=pts3d.astype(np.float32)
    pts2d=pts2d.astype(np.float32)

    dist_coef = np.zeros(4)
    #Change in OpenCV 3.1: 'strange' reshaping is required ( [N,3] to [N,1,3], and [N,2] to [N,1,2])
    #Note1: the reprojectionError must be set to "2" instead of default 8 value
    #reprojectionError: The parameter value is the maximum allowed distance between the observed and computed point projections to consider it an inlier
    #In french: seuil (distance) en dessous de laquelle les points peuvent etre "inliers".
    #Note2: additional return value '_' (versus opencv 2.4)
    
    pts3d_tmp=pts3d.reshape(-1,1,3)
    pts2d_tmp=pts2d.reshape(-1,1,2)
    _,rvec,tvec,inliers=cv2.solvePnPRansac(pts3d_tmp, pts2d_tmp, K, dist_coef,reprojectionError=2.0)

    R,b=cv2.Rodrigues(rvec)
    M_ext=np.hstack((R,tvec))
    #print("Rotation mat (R):",R)
    #print("Rotation vect (rad):",rvec)
    if verbose:
        print ("Rotation x (deg): " , np.round(rvec[0]*360.0/(2*np.pi),2))
        print ("Rotation y (deg): " , np.round(rvec[1]*360.0/(2*np.pi),2))
        print ("Rotation z (deg): " , np.round(rvec[2]*360.0/(2*np.pi),2))
        print ("Translation: ",np.round(tvec[0],2), ", " , np.round(tvec[1],2), ", "  , np.round(tvec[2],2))
        
        np.set_printoptions(suppress=True)
        print( "Matrice:\n", np.round(M_ext,2))
    return M_ext



def keypoints2numpyarray(kp):
    nb_points_ref=len(kp)
    points_ref=np.zeros((nb_points_ref,2),dtype=np.float)
    for i in range(0,nb_points_ref):
        points_ref[i,:]=kp[i].pt
    return points_ref


